#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2017-12-26
#
# Monitor status of Cisco BGP Peers (IPv4 and IPv6)
#
# 2018-05-24: changed counters to 1/s
# 2018-05-25: a lot of code cleanup
#             packet name changed from cisco_bgp to cisco_bgp_peer
#             added support of more then one address family per peer thanks to support[at]heinlein-support[dot]de
#             (changed item from remoteip to remoteip+familyname, rewrite of parser, inventory and check function)
# 2018-05-27: changed scan function from  '.1.3.6.1.4.1.9.9.187.1.2.7.1.3.* to sysdecr contains cisco
# 2018-05-28: changed wato, added peer alias, state if not found, infotext values
# 2018-05-29: fixed longoutput (removed not configured)
# 2018-11-02: modified scanfunction (from "find 'cisco' =-1"  to "'cisco' in OID"
# 2019-18-02: added fix for empty values ("" instead of "0") sugested by lbarbier[at]arkane-studios[dot]com
# 2020-02-24: added workaround for missing cbgpPeer2AddrFamily (example L2VPN EVPN peers)
#             thanks to jonale82[at]gmail[dot]com
# 2020-03-02: changed handling of perfdata, add only data the are really there (not None, instead of setting them to 0)
# 2020-06-04: code cleanup --> changed isdigit test to try/except loop, changed peer.get test to try/except loop
# 2020-09-10: fixed typo in metrics file. FMS --> FSM, thanks to martin[dot]pechstein[at]posteo[dot]de
# 2021-03-27: rewrite for CMK2.0
# 2021-03-28: added warning for missing prefix limit/warn threshold
# 2021-07-08: fixed handling of missing prefix limit/accepted prefixes/warn threshold.
#             thanks to Florian[dot]Dille[at]<redacted>>[dot]com
# 2021-07-13: fixed handling of perfdata (in/out updates, in/out total messages).
#             thanks to Florian[dot]Dille[at]<redacted>>[dot]com and timhor[at]outlook[dot]com
#             added type hinting for parsed data (section)
#             optimised data handling in check function (add prefixes to peer info)
#             made default state if item not found configurable via WATO
# 2021-07-21: fixed missing oid_end in parse function, thanks to oliverj@cmkforum
# 2021-08-18: fixed counter rollover, changed detect section to find only BGP enabled devices.
# 2012-08-20: cleanup, optimized parse function
# 2021-08-22: added WATO option for admin state down
# 2021-08-23: changed service name from "Cisco BGP peer" to "BGP peer" --> streamline with bgp_peer
#             get localas from BGP-4-MIB::bgpLocalAs if CISCO-BGP4-MIB::cbgpPeer2LocalAs not set
# 2021-08-29: fixed bgpLocalAs in BGP-4-MIB not set
# 2021-11-08: moved helper functions to utils/bgp_peer
#             moved longoutput/details to inventory based on inv_bgp_beer
#             merged WATO with bgp_peer
#             merged METRICS with bgp_peer
# 2021-11-13: fix for mixed peers (with and without prefixes on one host)
# 2021-11-14: merged check function with bgp_peer
#             merged parse function with bgp_peer
#             added basic support for VRFs (cbgpPeer3Entry)
# 2022.04.22: moved inventory plugin to inv_cisco_bgp_peer.py
# 2022-04-29: added info if device is admin prefix limit capable (device_admin_limit)
# 2022-04-30: code cleanup/streamlining
# 2022-05-09: added item to BgpPeer class, this is used in the discovery function
# 2022-05-11: changed bgp_get_peer_entry to get proper parameters instead of Nontransparent list
#             added remote_as to BgpPeerItem
# 2024-07-19: added in peerv2 local/remote AS/ID/Address for BGP topology
# 2024-12-20: added local address/as/identifier in cisco_bgp_peer for bgp_topology
# 2024-12-23: added local address/as/identifier in cisco_bgp_peer_3 for bgp_topology
#             fixed item in cisco_bgp_peer_3

from dataclasses import dataclass
from typing import List, Dict, Optional, Tuple
import copy

from cmk.agent_based.v2 import (
    OIDBytes,
    OIDEnd,
    SNMPSection,
    SNMPTree,
    StringByteTable,
    all_of,
    contains,
    exists,
)

from cmk_addons.plugins.bgp_peer.lib.bgp_peer import (
    BgpPeer,
    bgp_get_ip_address_from_oid,
    bgp_get_peer_entry,
    bgp_render_ip_address,
)


###########################################################################
#
#  cisco_bgp_peer (CISCO-BGP4-MIB:cbgpPeer2Entry)
#
###########################################################################


@dataclass
class CiscoPrefixes:
    address_family: str
    metric_count: List[Tuple[str, int]]
    metric_rate: List[Tuple[str, int]]
    accepted_prefixes: Optional[int]
    prefix_admin_limit: Optional[int]
    prefix_clear_threshold: Optional[int]
    prefix_threshold: Optional[int]


def parse_cisco_bgp_peer(string_table: List[StringByteTable]) -> Dict[str, BgpPeer]:
    peer2_entry, peer2_addr_family, base_local_as = string_table

    try:
        base_local_as = base_local_as[0][0]
    except (ValueError, IndexError):
        base_local_as = None

    peer_prefixes = {}
    # create dictionary by peer address and list of entries per address family
    for entry in peer2_addr_family:
        (
            oid_end,
            addr_family_name,
            accepted_prefixes,
            denied_prefixes,
            prefix_admin_limit,
            prefix_threshold,
            prefix_clear_threshold,
            advertised_prefixes,
            suppressed_prefixes,
            withdrawn_prefixes
        ) = entry

        if not oid_end == '':
            remote_addr = bgp_get_ip_address_from_oid(oid_end)
            prefixes = CiscoPrefixes(
                address_family=addr_family_name,
                metric_count=[],
                metric_rate=[],
                accepted_prefixes=int(accepted_prefixes) if accepted_prefixes.isdigit() else None,
                prefix_admin_limit=int(prefix_admin_limit) if prefix_admin_limit.isdigit() else None,
                prefix_clear_threshold=int(prefix_threshold) if prefix_threshold.isdigit() else None,
                prefix_threshold=int(prefix_threshold) if prefix_threshold.isdigit() else None
            )
            for key, value in [
                ('advertisedprefixes', advertised_prefixes),
                ('deniedprefixes', denied_prefixes),
                ('suppressedprefixes', suppressed_prefixes),
                ('withdrawnprefixes', withdrawn_prefixes),
            ]:
                try:
                    prefixes.metric_rate.append((key, int(value)))
                except ValueError:
                    pass

            if peer_prefixes.get(remote_addr):
                peer_prefixes[remote_addr].append(prefixes)
            else:
                peer_prefixes[remote_addr] = [prefixes]

    # create dictionary from cbgpPeer2Entry (peer ip address as index)
    peer_table: Dict[str, BgpPeer] = {}
    for entry in peer2_entry:
        (
            remote_addr,
            peer_state,
            admin_state,
            local_addr,
            local_as,
            local_id,
            remote_as,
            remote_id,
            in_updates,
            out_updates,
            in_messages,
            out_messages,
            fsm_established_transitions,
            fsm_established_time,
            in_update_elapsed_time
        ) = entry

        if len(local_addr) == 4:
            local_addr_type = '1'  # IPv4
        elif len(local_addr) == 16:
            local_addr_type = '2'  # IPv6
        else:
            local_addr_type = '0'

        remote_addr = bgp_get_ip_address_from_oid(remote_addr)
        bgp_peer = bgp_get_peer_entry(
            admin_state=admin_state,
            fsm_established_time=fsm_established_time,
            fsm_established_transitions=fsm_established_transitions,
            in_messages=in_messages,
            in_update_elapsed_time=in_update_elapsed_time,
            in_updates=in_updates,
            local_addr=bgp_render_ip_address(local_addr_type, local_addr),
            local_as=local_as if local_as != '0' else base_local_as,
            local_id=local_id,
            out_messages=out_messages,
            out_updates=out_updates,
            peer_state=peer_state,
            remote_addr=remote_addr,
            remote_as=remote_as,
            remote_id=remote_id,
        )
        if bgp_peer:
            prefixes = peer_prefixes.get(remote_addr, [])
            if prefixes:
                for address_family in prefixes:
                    item = f'{remote_addr} {address_family.address_family}'
                    temp_bgp_peer = copy.deepcopy(bgp_peer)
                    peer_table[item] = temp_bgp_peer[remote_addr]
                    peer_table[item].accepted_prefixes = address_family.accepted_prefixes
                    peer_table[item].prefix_admin_limit = address_family.prefix_admin_limit
                    peer_table[item].prefix_clear_threshold = address_family.prefix_clear_threshold
                    peer_table[item].prefix_threshold = address_family.prefix_threshold
                    peer_table[item].device_admin_limit = True
                    peer_table[item].metric_rate += address_family.metric_count
                    peer_table[item].metric_rate += address_family.metric_rate
                    peer_table[item].item['address_family'] = address_family.address_family
            else:
                peer_table.update(bgp_peer)

    return peer_table


snmp_section_cisco_bgp_peer = SNMPSection(
    name='cisco_bgp_peer',
    parse_function=parse_cisco_bgp_peer,
    parsed_section_name='bgp_peer',
    supersedes=['bgp_peer', 'cisco_bgp_peerv2'],
    fetch=[
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.187.1.2.5.1',  # CISCO-BGP4-MIB::cbgpPeer2Entry
            oids=[
                OIDEnd(),
                '3',  # cbgpPeer2State
                '4',  # cbgpPeer2AdminStatus
                OIDBytes('6'),  # cbgpPeer2LocalAddr
                '8',  # cbgpPeer2LocalAs -> empty
                '9',  # cbgpPeer2LocalIdentifier
                '11',  # cbgpPeer2RemoteAs
                '12',  # cbgpPeer2RemoteIdentifier
                '13',  # cbgpPeer2InUpdates
                '14',  # cbgpPeer2OutUpdates
                '15',  # cbgpPeer2InTotalMessages
                '16',  # cbgpPeer2OutTotalMessages
                '18',  # cbgpPeer2FsmEstablishedTransitions
                '19',  # cbgpPeer2FsmEstablishedTime
                '27',  # cbgpPeer2InUpdateElapsedTime
            ]
        ),
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.187.1.2',  # cbgpPeer
            oids=[
                OIDEnd(),  #
                # .7.1 --> cbgpPeer2AddrFamilyEntry
                '7.1.3',  # cbgpPeer2AddrFamilyName
                # .8.1 --> cbgpPeer2AddrFamilyPrefixEntry
                '8.1.1',  # cbgpPeer2AcceptedPrefixes
                '8.1.2',  # cbgpPeer2DeniedPrefixes
                '8.1.3',  # cbgpPeer2PrefixAdminLimit
                '8.1.4',  # cbgpPeer2PrefixThreshold
                '8.1.5',  # cbgpPeer2PrefixClearThreshold
                '8.1.6',  # cbgpPeer2AdvertisedPrefixes
                '8.1.7',  # cbgpPeer2SuppressedPrefixes
                '8.1.8',  # cbgpPeer2WithdrawnPrefixes
            ]
        ),
        SNMPTree(
            base='.1.3.6.1.2.1.15',  # BGP-4-MIB
            oids=[
                '2',  # bgpLocalAs
            ]
        )
    ],
    detect=all_of(
        contains('.1.3.6.1.2.1.1.1.0', 'Cisco'),
        exists('.1.3.6.1.4.1.9.9.187.1.2.5.1.3.*')
    ),
)


###########################################################################
#
#  cisco_bgp_peer (CISCO-BGP4-MIB:cbgpPeer3Entry)
#
###########################################################################


# ToDo: add support for non IPv4, needs sample
def parse_cisco_bgp_peer_3(string_table: List[StringByteTable]) -> Dict[str, BgpPeer]:
    peer_table = {}
    for entry in string_table[0]:
        (
            remote_addr_type,
            remote_addr,
            routing_instance,
            peer_state,
            admin_state,
            local_addr,
            local_as,
            local_id,
            remote_as,
            remote_id,
            in_updates,
            out_updates,
            in_messages,
            out_messages,
            fsm_established_transitions,
            fsm_established_time,
            in_update_elapsed_time
        ) = entry

        if remote_addr_type == '1':  # IPv4 address
            remote_addr = '.'.join(f'{m}' for m in remote_addr)
            local_addr = '.'.join(f'{m}' for m in local_addr)

            if routing_instance != 'default':
                routing_instance = f'vrf {routing_instance}'
            else:
                routing_instance = None

            bgp_peer = bgp_get_peer_entry(
                admin_state=admin_state,
                fsm_established_time=fsm_established_time,
                fsm_established_transitions=fsm_established_transitions,
                in_messages=in_messages,
                in_update_elapsed_time=in_update_elapsed_time,
                in_updates=in_updates,
                out_messages=out_messages,
                out_updates=out_updates,
                peer_state=peer_state,
                remote_addr=remote_addr,
                remote_as=remote_as,
                remote_id=remote_id,
                local_addr=local_addr,
                local_as=local_as,
                local_id=local_id,
            )
            if bgp_peer:
                if routing_instance:
                    bgp_peer[remote_addr].item['routing_instance'] = routing_instance
                    item = f'{remote_addr} {routing_instance}'
                    bgp_peer = {item: bgp_peer[remote_addr]}
                peer_table.update(bgp_peer)
    return peer_table


snmp_section_cisco_bgp_peer_3 = SNMPSection(
    name='cisco_bgp_peer_3',
    parse_function=parse_cisco_bgp_peer_3,
    parsed_section_name='bgp_peer',
    supersedes=['bgp_peer', 'cisco_bgp_peer', 'cisco_bgp_peerv2', 'cisco_bgp_peerv3'],
    fetch=[SNMPTree(
        base='.1.3.6.1.4.1.9.9.187.1.2.9.1',  # CISCO-BGP4-MIB::cbgpPeer3Entry
        oids=[
            '2',  # cbgpPeer3Type
            OIDBytes('3'),  # cbgpPeer3RemoteAddr
            '4',  # cbgpPeer3VrfName
            '5',  # cbgpPeer3State
            '6',  # cbgpPeer3AdminStatus
            OIDBytes('8'),  # cbgpPeer3LocalAddr
            '10',  # cbgpPeer3LocalAs
            '11',  # cbgpPeer3LocalIdentifier
            '13',  # cbgpPeer3RemoteAs
            '14',  # cbgpPeer3RemoteIdentifier
            '15',  # cbgpPeer3InUpdates
            '16',  # cbgpPeer3OutUpdates
            '17',  # cbgpPeer3InTotalMessages
            '18',  # cbgpPeer3OutTotalMessages
            '20',  # cbgpPeer3FsmEstablishedTransitions
            '21',  # cbgpPeer3FsmEstablishedTime
            '29',  # cbgpPeer3InUpdateElapsedTime
        ]
    )],
    detect=all_of(
        contains('.1.3.6.1.2.1.1.1.0', 'Cisco'),
        exists('.1.3.6.1.4.1.9.9.187.1.2.9.1.3.*')
    ),
)

# '1',  # cbgpPeer3VrfId
# '2',  # cbgpPeer3Type
# '3',  # cbgpPeer3RemoteAddr
# '4',  # cbgpPeer3VrfName
# '5',  # cbgpPeer3State
# '6',  # cbgpPeer3AdminStatus
# '7',  # cbgpPeer3NegotiatedVersion
# '8',  # cbgpPeer3LocalAddr
# '9',  # cbgpPeer3LocalPort
# '10',  # cbgpPeer3LocalAs
# '11',  # cbgpPeer3LocalIdentifier
# '12',  # cbgpPeer3RemotePort
# '13',  # cbgpPeer3RemoteAs
# '14',  # cbgpPeer3RemoteIdentifier
# '19',  # cbgpPeer3LastError
# '22',  # cbgpPeer3ConnectRetryInterval
# '23',  # cbgpPeer3HoldTime
# '24',  # cbgpPeer3KeepAlive
# '25',  # cbgpPeer3HoldTimeConfigured
# '26',  # cbgpPeer3KeepAliveConfigured
# '27',  # cbgpPeer3MinASOriginationInterval
# '28',  # cbgpPeer3MinRouteAdvertisementInterval
# '30',  # cbgpPeer3LastErrorTxt
# '31',  # cbgpPeer3PrevState
